// $Id: cvsd.cpp,v 1.2 2000/06/12 00:24:40 flip Exp $
/*
   cvsd: a little chroot/suid wrapper to run CVS more securely.

   Chris Black <cblack@mokey.com> http://cblack.mokey.com/

   Many ideas taken from various CVS documentation and discussions and 
   basic chroot wrapper scripts.
   This code is under the GPL, see www.gnu.org for details.
   No warranty, I take no responsibility if this allows someone to put a 
   subtle bug into your missile tracking software during development which 
   leads to Bad Things.

   Philippe Kehl <phkehl@gmx.net> http://guv.ethz.ch/~flip/

   some ideas and pieces of code i took from ....'s excellent bk2site

   improved version. cvsd 0.7
   - cleaner config file and method to read it

*/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include<iostream>
#include<stdio.h>
//#include<stdlib.h>
#include<cstdlib>
#include<string>
//#include<iomanip>
#include<fstream>
#include<vector>
//#include <stdio.h>
//#include <string.h>
//#include <pwd.h>
//#include <sys/types.h>
//#include <sys/socket.h>
#include <unistd.h>
//#include <errno.h>
#include <syslog.h>
//#include <string>
//#include <fstream>

using namespace std;

const string CONFIGFILE = "/etc/cvsd.conf";
const int NUMVARS = 4;
const string varNames[NUMVARS] = {
  /*1*/ "CVSDUID", /*2*/ "CVSDGID", /*3*/ "CVSDROOT", /*4*/ "REPOSITORY"
};
const int CVSDUID = 0;
const int CVSDGID = 1;
const int CVSDROOT = 2;
const int REPOSITORY = 3;


int cmpNoCase(const string& s, const string &s2)
{
  string::const_iterator p = s.begin();
  string::const_iterator p2 = s2.begin();  
  while (p != s.end() && p2 != s2.end()) {
    if (toupper(*p) != toupper(*p2))
      return (toupper(*p) < toupper(*p2)) ? -1 : 1;
    ++p;
    ++p2;
  }
  return (s2.size() == s.size()) ? 0 : (s.size() < s2.size()) ? -1 : 1;
}
int replaceAll(string & s, string s1, string s2)
{
  int c = 0;
  string::size_type f1 = 0;
  while ((f1 = s.find(s1,f1)) != string::npos) {
    c++;
    s.replace(f1,s1.size(), s2);
    f1 += s2.size();
  }
  return c;
}
string trim(const string & s) 
{
  int l = s.length();
  int left =0, right =l-1;
  for (; left <= right; left++) 
    if (!(isspace(s[left]))) break;
  for (; right >= left; right--)
    if (!(isspace(s[right]))) break;
  string t(s.substr(left, right+1-left));
  replaceAll(t, "\n", "");
  return t;
}

int main(int argc, char *argv[], char *env[]) {
  /*  extern int errno; */
  /*  extern char **environ; */
  /*  int i; char *buf, *p, *cvsroot; */

  // open system logger...

  openlog("cvsd", LOG_PID, LOG_DAEMON);
  syslog(LOG_INFO, "Connection made.\n");

  // set default values...

  string varValues[NUMVARS];
  varValues[CVSDUID] = "njet";
  varValues[CVSDGID] = "njet";
  varValues[CVSDROOT] = "njet";
  varValues[REPOSITORY] = "njet";

  vector<string> repositories;

  // read in config file..

  ifstream configStream(CONFIGFILE.c_str(), ios::in);
  if (!configStream) {
    syslog(LOG_INFO, "Cannot read config file %s.", CONFIGFILE.c_str()); 
    exit(1);
  } else {
    string curVarName;
    string curVarValue;
    while (configStream >> curVarName) {
      getline(configStream, curVarValue);
      curVarValue = trim(curVarValue);      
      if (curVarName[0] == '#') continue;

      if (curVarValue[0] == '"') //get rid of surrounding "'s
      	curVarValue.erase(0,1);
      if (curVarValue[curVarValue.size() -1] == '"')
      	curVarValue.erase(curVarValue.size() - 1,1);

      int i, found = 0;
      for (i = 0; i < NUMVARS; i++){
     	if (cmpNoCase(varNames[i],curVarName) == 0){
     	  varValues[i] = curVarValue;
     	  found = 1;
     	  break;
     	}
      }
      // REPOSITORY is special, since it can appear many times.
      if (cmpNoCase(varNames[REPOSITORY], curVarName) == 0){
	repositories.push_back(curVarValue);
	found = 1;
      }
      // if (found)
      //   syslog(LOG_INFO, "key = \"%s\", value = \"%s\"", curVarName.c_str(), curVarValue.c_str());
      // else
      if (!found)
	syslog(LOG_INFO, "unknow config key %s", curVarName.c_str());
    }

    // now see what we have...

    syslog(LOG_INFO, "CVSDUID = %s, CVSDGID = %s" , varValues[CVSDUID].c_str(), varValues[CVSDGID].c_str()); 
    syslog(LOG_INFO, "CVSDROOT = %s", varValues[CVSDROOT].c_str());
    for (unsigned int nr = 0; nr < repositories.size(); nr++)
      syslog(LOG_INFO, "REPOSITORY[%i] = \"%s\"", nr, repositories[nr].c_str());

    // build arguments array...

    char * args[repositories.size()+2];

    args[0] = "cvs";
    unsigned int nr = 0;
    for (nr; nr < repositories.size(); nr++) {
      char meier[255] = "--allow-root=";
      //      syslog(LOG_INFO, "%i - %s", nr, repositories[nr].c_str());
      strcat(meier, repositories[nr].c_str());
      args[nr+1] = (char *)malloc((strlen((char *)meier)) * sizeof(char));
      strcpy(args[nr+1], meier);
    }
    //args[nr+1] = "-t";
    args[nr+1] = "pserver";
    args[nr+2] = NULL;

    for(int i  = 0; args[i] != NULL; i++) {
      syslog(LOG_INFO, "cvs args[%d] = \"%s\"\n", i, args[i]);
    } 

    // now going for the real thing...

    if(chdir( varValues[CVSDROOT].c_str() )) {
      syslog(LOG_ERR, "Could not chdir to %s\n", varValues[CVSDROOT].c_str()); exit(2); }
    
    if(chroot( varValues[CVSDROOT].c_str() )) {
      syslog(LOG_ERR, "Could not chroot to %s\n", varValues[CVSDROOT].c_str()); exit(3); }
    
    int uid = atoi(varValues[CVSDUID].c_str());
    int gid = atoi(varValues[CVSDGID].c_str());
	   
    if (uid == 0) { syslog(LOG_INFO, "uid 0 not allowed"); exit(4); }
    if (gid == 0) { syslog(LOG_INFO, "gid 0 not allowed"); exit(4); }
    
    if (setgid( gid )) {
      syslog(LOG_ERR, "Could not setgid to %d\n", gid ); exit(5); }
    
    if (setuid(uid)) {
      syslog(LOG_ERR, "Could not setuid to %d\n", uid ); exit(5); }
    
    chdir("/");

    if (setenv("HOME", "/", 1) != 0) {
      syslog(LOG_ERR, "Could not set env var HOME.\n"); exit(6); }

    if (setenv("TMPDIR", "/tmp", 1) != 0) {
      syslog(LOG_ERR, "Could not set env var TMPDIR.\n"); exit(6); }

    if (setenv("CVSUMASK", "027", 1) != 0) {
      syslog(LOG_ERR, "Could not set env var CVSUMASK.\n"); exit(6); }
    
    
  
    syslog(LOG_INFO, "Now passing task to cvs...");
    
    int ret = execvp("/bin/cvs", args);
    
    syslog(LOG_ERR, "Error executing cvs (%i)", ret);
    
    exit(7);
    
    
    
  }
  
  

}





